from datetime import datetime
from typing import Optional, List

from sqlalchemy.orm.decl_api import DeclarativeBase
from sqlalchemy import String, Integer, Boolean, DateTime, Index, ForeignKey, Text
from sqlalchemy.orm.decl_api import Mapped
from sqlalchemy.orm.decl_api import mapped_column
from sqlalchemy.orm import relationship


class Base(DeclarativeBase):
    pass


class Patch(Base):
    __tablename__ = 'patches'
    release: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    kernel_id: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    patch_level: Mapped[int] = mapped_column(Integer, nullable=False, primary_key=True)
    version: Mapped[int] = mapped_column(Integer, nullable=False)


Index('idx_patches_kernel_id_patch_level', Patch.kernel_id, Patch.patch_level.desc())


class Fixup(Base):
    __tablename__ = 'fixups'
    release: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    kernel_id: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    fixup_for: Mapped[str] = mapped_column(Integer, nullable=False, primary_key=True)
    fixup_from: Mapped[int] = mapped_column(Integer, nullable=False)


Index('idx_fixups_kernel_for_from', Fixup.kernel_id, Fixup.fixup_for, Fixup.fixup_from)


class Release(Base):
    __tablename__ = 'releases'
    unit_id: Mapped[int] = mapped_column(Integer, nullable=False, default=0, server_default='0', primary_key=True)
    stream: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    release: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    feed: Mapped[str] = mapped_column(String, nullable=False, primary_key=True)
    deployed: Mapped[datetime] = mapped_column(DateTime, nullable=False)
    enabled: Mapped[bool] = mapped_column(Boolean, nullable=False)


Index('idx_releases_stream_release', Release.stream, Release.release)


class UpdateInfo(Base):
    __tablename__ = "update_info"

    params_hash: Mapped[str] = mapped_column(String(64), primary_key=True)
    params_str: Mapped[str] = mapped_column(Text, nullable=False)
    fetched_at: Mapped[datetime] = mapped_column(default=datetime.utcnow, nullable=False)
    blocked: Mapped[str] = mapped_column(Text, nullable=False)
    error: Mapped[Optional[str]] = mapped_column(String, nullable=True)

    releases_meta: Mapped[List["RemoteReleaseMeta"]] = relationship(
        "RemoteReleaseMeta",
        cascade="all, delete-orphan",
    )


class RemoteReleaseMeta(Base):
    __tablename__ = "remote_release_meta"

    update_info_params_hash: Mapped[str] = mapped_column(
        ForeignKey("update_info.params_hash", ondelete="CASCADE"),
        primary_key=True,
    )
    release: Mapped[str] = mapped_column(String(64), primary_key=True)
    hash: Mapped[str] = mapped_column(String(64), nullable=False)
    ps_deployed: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))
    ps_released: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))

    # passive_delete is needed to skip related release_distros entries deletion
    # https://stackoverflow.com/a/71101140
    distros: Mapped[List["ReleaseDistro"]] = relationship(
        "ReleaseDistro",
        primaryjoin="RemoteReleaseMeta.release == foreign(ReleaseDistro.release)",
        passive_deletes="all",
    )

    def _format_dt(self, dt: Optional[datetime]) -> str:
        return dt.strftime('%Y-%m-%d %H:%M:%S.%f') if dt else ""

    @property
    def ps_deployed_str(self) -> str:
        return self._format_dt(self.ps_deployed)

    @property
    def ps_released_str(self) -> str:
        return self._format_dt(self.ps_released)


class ReleaseDistro(Base):
    __tablename__ = "release_distros"

    release: Mapped[str] = mapped_column(String(64), primary_key=True)
    distro: Mapped[str] = mapped_column(String(64), primary_key=True)
