#!/usr/bin/env python3
from __future__ import annotations

import os
import os.path
import sys
from typing import cast

PRODUCT_DIR = '/usr/share/kcare-eportal'
python_ver = '{}.{}'.format(*sys.version_info[:2])
sys.path.insert(0, f'{PRODUCT_DIR}/vendor/lib/python{python_ver}/site-packages')
sys.path.insert(0, f'{PRODUCT_DIR}/vendor/lib64/python{python_ver}/site-packages')
sys.path.insert(0, PRODUCT_DIR)

import glob
import logging.config
import runpy
import shutil

import config
from alembic import command
from alembic.config import Config as AlembicConfig
from alembic.script import Script, ScriptDirectory

from eportal import utils
from eportal import EPORTAL_VERSION
from eportal.patches import session
from eportal.dbutils import transaction

ALEMBIC_CONFIG_PATH = os.path.join(PRODUCT_DIR, 'alembic.ini')
logger = logging.getLogger()


class SchemaMigrator:
    CONFIG_SECTIONS = ('alembic', 'patches-data', 'spooler')

    def __init__(self, config_path: str) -> None:
        self.configs = {section: AlembicConfig(config_path, ini_section=section) for section in self.CONFIG_SECTIONS}

    def upgrade(self) -> None:
        """Upgrade DB to 'head' revision"""
        for section, cfg in self.configs.items():
            logger.warning(f'[{section}] Upgrading to revision "head"')
            command.upgrade(cfg, 'head')

    def get_head_revision(self, section: str) -> Script:
        script = ScriptDirectory.from_config(self.configs[section])
        revs = script.get_revisions(script.get_heads())
        if len(revs) == 1:
            return revs[0]
        branch = 'unit' if config.BUNITS else 'main'
        return next(filter(lambda r: branch in r.branch_labels, revs))

    def get_revision(self, section: str, revision: str) -> Script | None:
        try:
            script = ScriptDirectory.from_config(self.configs[section])
            return script.get_revision(revision)
        except Exception:
            return None

    def get_head_revisions(self) -> dict[str, Script]:
        return {section: self.get_head_revision(section) for section in self.CONFIG_SECTIONS}

    def get_branch_point(self, section: str, start_revision: Script | None = None) -> Script | None:
        script = ScriptDirectory.from_config(self.configs[section])
        revision = start_revision if start_revision else self.get_head_revision(section)
        if revision.branch_labels and len(revision.branch_labels) > 1:
            logger.error(f'[{section}] Revision {revision} is after merge point')
            return None
        while revision and not revision.is_branch_point:
            revision = script.get_revision(cast(str, revision.down_revision))
        return revision

    def downgrade(self, revisions: dict[str, Script]) -> None:
        """Downgrade DB to revision provided"""
        for section, cfg in self.configs.items():
            if section not in revisions:
                continue
            target_revision = revisions[section]
            logger.warning(f'[{section}] Downgrading to revision {target_revision}')
            command.downgrade(cfg, target_revision.revision)
            if target_revision.branch_labels == {'main'}:
                # we're on the main branch, bunit-related schema should be removed
                branchpoint = self.get_branch_point(section, target_revision)
                if not branchpoint:
                    logger.error(f'[{section}] Branchpoint not found for revision {target_revision}')
                else:
                    logger.warning(f'[{section}] Downgrading "unit" branch to revision {branchpoint}')
                    command.downgrade(cfg, f'unit@{branchpoint.revision}')


def get_sqlite_file_pathes() -> list[str]:
    return glob.glob(os.path.join(config.BASE_DIR, '*.sqlite*'))


def create_oplog() -> None:
    # do these import after become_owner to be sure all databases would be
    # created with proper owner
    from eportal import oplog

    logger.info('Creating oplog table ... ')
    oplog.create_table()
    verify_db_owner(os.geteuid())
    logger.info('done.')


def upgrade() -> None:
    current_schema = SchemaMigrator(ALEMBIC_CONFIG_PATH)
    current_schema.upgrade()


def downgrade_if_required(migration_dir: str) -> None:
    """
    Downgrade is called on alembic scripts preserved from previous
    (more recent) package at :param migration_dir, since we don't have required
    schema scripts and migration logic in the installed (older) eportal version.
    The target revision for the downgrade corresponds to the "head" revision
    from the installed package.

    To downgrade to eportal version <= 2.11-1 preserved script
    {config.BASE_DIR}/migration/migratedb.py should be called explicitly
    from postrm/postun package triggers or manually in case docker container delivery
    """

    preserved_schema = SchemaMigrator(os.path.join(migration_dir, 'alembic.ini'))
    current_schema = SchemaMigrator(ALEMBIC_CONFIG_PATH)
    target_revisions = current_schema.get_head_revisions()
    preserved_revisions = [preserved_schema.get_revision(s, r.revision) for s, r in target_revisions.items()]
    if all(preserved_revisions) and any(not r.is_head for r in preserved_revisions if r):
        preserved_schema.downgrade(target_revisions)
    else:
        logger.info(f'Downgrade is not required for revisions: {target_revisions}')


def get_file_owner(path: str) -> tuple[int, int]:
    file_stat = os.stat(path)
    return file_stat.st_uid, file_stat.st_gid


def verify_db_owner(proc_euid: int) -> None:
    path_with_owner = [(path, *get_file_owner(path)) for path in [config.BASE_DIR] + get_sqlite_file_pathes()]
    mismatch = {path: uid for path, uid, _ in path_with_owner if uid != proc_euid}
    if mismatch:
        logger.error(f'File owner mismatch found for {{path: uid}}: {mismatch}. Expected uid: {proc_euid}')
        sys.exit(1)


def preserve_current_migration_scripts(migration_dir: str) -> None:
    if os.path.isdir(migration_dir):
        shutil.rmtree(migration_dir)
    os.makedirs(migration_dir, mode=0o700)
    shutil.copy(__file__, migration_dir)
    shutil.copy(os.path.join(PRODUCT_DIR, 'alembic.ini'), migration_dir)
    shutil.copytree(os.path.join(PRODUCT_DIR, 'migrations'), os.path.join(migration_dir, 'migrations'))
    utils.atomic_write(os.path.join(migration_dir, 'version'), content=EPORTAL_VERSION)


@transaction(session)
def load_release_distros_for_deployed_releases() -> None:
    # import here to avoid circular imports
    from eportal import patches_constants as pc, patches_impl
    from eportal.patches import api as patches_api, models as mp

    if session.query(mp.ReleaseDistro).first():
        # we have already stored distros, no need to run this again
        return

    for stream in pc.Stream:
        # Assume that locally deployed releases were deployed for BU=0 because CLI doesn't support BU
        # In case of BU enabled we assume that releases were downloaded from patchserver
        # and release_distros will be populated during the next updates fetch
        patch_store = patches_impl.get_store(0, stream)
        logger.warning('saving release:distro pairs for %s stream', stream)
        releases = session.query(mp.Release.release).distinct().filter(mp.Release.stream == stream.value)
        for release in releases:
            distros = list(patch_store.release_flavors(release.release))
            patches_api.insert_release_distros(release.release, distros)


def main() -> None:
    if os.geteuid() == 0:
        # fix db permissions
        owner = utils.get_base_owner()
        broken_owner_files = [f for f in get_sqlite_file_pathes() if get_file_owner(f) != owner]
        for f in broken_owner_files:
            logger.warning(f'Changing owner for file {f}: {get_file_owner(f)} -> {owner}')
            shutil.chown(f, *owner)

    utils.become_owner()
    verify_db_owner(os.geteuid())
    create_oplog()

    current_migration_dir = os.path.dirname(os.path.abspath(__file__))
    if current_migration_dir == PRODUCT_DIR:
        preserved_migration_dir = os.path.join(config.BASE_DIR, 'migration')
        # First, call migration script from previous package for possible downgrade
        migration_script = os.path.join(preserved_migration_dir, os.path.basename(__file__))
        if os.path.exists(migration_script):
            try:
                runpy.run_path(migration_script, run_name='__main__')
            except Exception as err:
                logger.error("Migration script from previous package failed: %r", err)
        upgrade()
        preserve_current_migration_scripts(preserved_migration_dir)

        try:
            load_release_distros_for_deployed_releases()
        except Exception:
            logger.exception('failed to save release:distro pairs to DB')
    else:
        downgrade_if_required(current_migration_dir)


if __name__ in ('__main__', '<run_path>'):
    logging.config.fileConfig(ALEMBIC_CONFIG_PATH)
    main()
